#   Copyright (C) 2016 Canonical Ltd.
#
#   Author: Ryan Harper <ryan.harper@canonical.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import datetime
import json
import sys
import time
from typing import IO, Any, Dict, List, Optional, Tuple, Union, cast

from cloudinit import subp, util
from cloudinit.distros import uses_systemd

# Example events:
#     {
#             "description": "executing late commands",
#             "event_type": "start",
#             "level": "INFO",
#             "name": "cmd-install/stage-late"
#             "origin": "cloudinit",
#             "timestamp": 1461164249.1590767,
#     }
#     {
#         "description": "executing late commands",
#         "event_type": "finish",
#         "level": "INFO",
#         "name": "cmd-install/stage-late",
#         "origin": "cloudinit",
#         "result": "SUCCESS",
#         "timestamp": 1461164249.1590767
#     }

format_key = {
    "%d": "delta",
    "%D": "description",
    "%E": "elapsed",
    "%e": "event_type",
    "%I": "indent",
    "%l": "level",
    "%n": "name",
    "%o": "origin",
    "%r": "result",
    "%t": "timestamp",
    "%T": "total_time",
}

SUCCESS_CODE = "successful"
FAIL_CODE = "failure"
CONTAINER_CODE = "container"
TIMESTAMP_UNKNOWN = (FAIL_CODE, -1, -1, -1)

Event = Dict


def format_record(msg: str, event: Event) -> str:
    for i, j in format_key.items():
        if i in msg:
            # ensure consistent formatting of time values
            if j in ["delta", "elapsed", "timestamp"]:
                msg = msg.replace(i, "{%s:08.5f}" % j)
            else:
                msg = msg.replace(i, "{%s}" % j)
    return msg.format(**event)


def event_name(event: Optional[Event]) -> Optional[str]:
    if event:
        return event.get("name")
    return None


def event_type(event: Optional[Event]) -> Optional[str]:
    if event:
        return event.get("event_type")
    return None


def event_parent(event: Optional[Event]) -> Optional[str]:
    name = event_name(event)
    if name:
        return name.split("/")[0]
    return None


def event_timestamp(event: Event) -> float:
    ts = event.get("timestamp")
    if ts is None:
        raise ValueError("Event is missing a 'timestamp'")
    return float(ts)


def event_datetime(event: Event) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(
        event_timestamp(event), datetime.timezone.utc
    )


def delta_seconds(t1: datetime.datetime, t2: datetime.datetime) -> float:
    return (t2 - t1).total_seconds()


def event_duration(start: Event, finish: Event) -> float:
    return delta_seconds(event_datetime(start), event_datetime(finish))


def event_record(
    start_time: datetime.datetime,
    start: Event,
    finish: Event,
) -> Event:
    record = finish.copy()
    name = event_name(start)
    indent = "|"
    if name:
        indent += " " * (name.count("/") - 1) + "`->"
    record.update(
        {
            "delta": event_duration(start, finish),
            "elapsed": delta_seconds(start_time, event_datetime(start)),
            "indent": indent,
        }
    )

    return record


def total_time_record(total_time: float) -> str:
    return "Total Time: %3.5f seconds\n" % total_time


class SystemctlReader:
    """
    Class for dealing with all systemctl subp calls in a consistent manner.
    """

    def __init__(self, property: str, parameter: Optional[str] = None):
        self.stdout: Optional[str] = None
        self.args: Any = [subp.which("systemctl"), "show"]

        if parameter:
            self.args.append(parameter)
        # --timestamp=utc is needed for native date strings. Otherwise,
        # the datetime will be returned in the local timezone (which would be
        # a problem for strptime used later on in this method)
        # This option does not affect monotonic properties (values as
        # microsecond int)
        self.args.extend(["-p", property, "--timestamp=us+utc"])

        # Don't want the init of our object to break. Instead of throwing
        # an exception, set an error code that gets checked when data is
        # requested from the object
        self.failure = self._subp()

    def _subp(self) -> Optional[Union[str, Exception]]:
        """
        Make a subp call based on set args and handle errors by setting
        failure code

        :return: whether the subp call failed or not
        """
        try:
            value, err = subp.subp(self.args, capture=True)
            if err:
                return err
            self.stdout = value
            return None
        except Exception as systemctl_fail:
            return systemctl_fail

    def convert_val_to_float(self) -> float:
        """
        If subp call succeeded, return the timestamp from subp as a float.

        :return: timestamp as a float
        """
        # subp has 2 ways to fail: it either fails and throws an exception,
        # or returns an error code. Raise an exception here in order to make
        # sure both scenarios throw exceptions
        if self.failure:
            raise RuntimeError(
                "Subprocess call to systemctl has failed, "
                "returning error code ({})".format(self.failure)
            )
        # this should never happen as the call to subp succeeded

        if self.stdout is None:
            raise RuntimeError(
                "stdout of subprocess call to systemctl is None"
            )

        # Output from systemctl show has the format Property=Value.

        val = self.stdout.split("=")[1].strip()

        if val.isnumeric():
            # Float Timestamps reported by systemctl are in
            # microseconds, converting to seconds
            # For example, UserspaceMonotonic=1929304
            timestamp = float(val) / 1000000
        else:
            # The format in this case is always "%a %Y-%m-%d %H:%M:%S %Z"
            # For example, UserspaceTimestamp=Wed 2025-07-30 05:14:32 UTC

            # strptime returns a naive datetime so we need to explicitly
            # set the timezone of this datetime
            # at the timezone of the parsed string (utc)
            timestamp = (
                datetime.datetime.strptime(val, "%a %Y-%m-%d %H:%M:%S.%f %Z")
                .replace(tzinfo=datetime.timezone.utc)
                .timestamp()
            )

        return timestamp


def dist_check_timestamp() -> Tuple[str, float, float, float]:
    """
    Determine which init system a particular linux distro is using.
    Each init system (systemd, etc) has a different way of
    providing timestamps.

    :return: timestamps of kernelboot, kernelendboot, and cloud-initstart
    or TIMESTAMP_UNKNOWN if the timestamps cannot be retrieved.
    """

    if uses_systemd():
        return gather_timestamps_using_systemd()

    # Use dmesg to get timestamps if the distro does not have systemd
    if util.is_FreeBSD() or "gentoo" in util.system_info()["system"].lower():
        return gather_timestamps_using_dmesg()

    # this distro doesn't fit anything that is supported by cloud-init. just
    # return error codes
    return TIMESTAMP_UNKNOWN


def gather_timestamps_using_dmesg() -> Tuple[str, float, float, float]:
    """
    Gather timestamps that corresponds to kernel begin initialization,
    kernel finish initialization using dmesg as opposed to systemctl

    :return: the two timestamps plus a dummy timestamp to keep consistency
    with gather_timestamps_using_systemd
    """
    try:
        data, _ = subp.subp(["dmesg"], capture=True)
        split_entries = data[0].splitlines()
        for i in split_entries:
            if i.decode("UTF-8").find("user") != -1:
                splitup = i.decode("UTF-8").split()
                stripped = splitup[1].strip("]")

                # kernel timestamp from dmesg is equal to 0,
                # with the userspace timestamp relative to it.
                user_space_timestamp = float(stripped)
                kernel_start = float(time.time()) - float(util.uptime())
                kernel_end = kernel_start + user_space_timestamp

                # systemd wont start cloud-init in this case,
                # so we cannot get that timestamp
                return SUCCESS_CODE, kernel_start, kernel_end, kernel_end

    except Exception:
        pass
    return TIMESTAMP_UNKNOWN


def gather_timestamps_using_systemd() -> Tuple[str, float, float, float]:
    """
    Gather timestamps that corresponds to kernel begin initialization,
    kernel finish initialization. and cloud-init systemd unit activation

    :return: the three timesread_propertystamps
    """
    try:
        # The use of the monotonic timestamps is needed in cloud-init-related
        # dates to account for the 2-second delay when cloud-init sets up NTP
        if util.is_container():
            # lxc based containers do not set their monotonic zero point to be
            # when the container starts,
            # instead keep using host boot as zero point
            kernel_start = SystemctlReader(
                "UserspaceTimestamp"
            ).convert_val_to_float()
            monotonic_offset = SystemctlReader(
                "UserspaceTimestampMonotonic"
            ).convert_val_to_float()
        else:
            kernel_start = SystemctlReader(
                "KernelTimestamp"
            ).convert_val_to_float()
            monotonic_offset = SystemctlReader(
                "KernelTimestampMonotonic"
            ).convert_val_to_float()
        kernel_end = (
            SystemctlReader(
                "UserspaceTimestampMonotonic"
            ).convert_val_to_float()
            - monotonic_offset
        )
        cloudinit_sysd = (
            SystemctlReader(
                "InactiveExitTimestampMonotonic", "cloud-init-local"
            ).convert_val_to_float()
            - monotonic_offset
        )

    except Exception as e:
        # Except ALL exceptions as Systemctl reader can throw many different
        # errors, but any failure in systemctl means that timestamps cannot be
        # obtained
        print(e)
        return TIMESTAMP_UNKNOWN

    status = CONTAINER_CODE if util.is_container() else SUCCESS_CODE

    return (
        status,
        kernel_start,
        kernel_start + kernel_end,
        kernel_start + cloudinit_sysd,
    )


def generate_records(
    events: List[Event],
    print_format: str = "(%n) %d seconds in %I%D",
) -> List[List[str]]:
    """
    Take in raw events and create parent-child dependencies between events
    in order to order events in chronological order.

    :param events: JSONs from dump that represents events taken from logs
    :param print_format: formatting to represent event, time stamp,
    and time taken by the event in one line

    :return: boot records ordered chronologically
    """

    sorted_events = sorted(events, key=lambda x: x["timestamp"])
    records: List[str] = []
    start_time = None
    total_time = 0.0
    stage_start_time = {}
    boot_records = []

    unprocessed = []
    for e in range(len(sorted_events)):
        event = events[e]
        try:
            next_evt = events[e + 1]
        except IndexError:
            next_evt = None

        if event_type(event) == "start":
            if records and event.get("name") == "init-local":
                records.append(total_time_record(total_time))
                boot_records.append(records)
                records = []
                start_time = None
                total_time = 0.0

            if start_time is None:
                start_time = event_datetime(event)
                stage_start_time[event_parent(event)] = start_time

            # see if we have a pair
            if event_name(event) == event_name(next_evt):
                if event_type(next_evt) == "finish":
                    # next_evt is a dictionary because event_type has extracted
                    # an event type from it:
                    next_evt = cast(Dict, next_evt)
                    records.append(
                        format_record(
                            print_format,
                            event_record(start_time, event, next_evt),
                        )
                    )
            else:
                # This is a parent event
                records.append("Starting stage: %s" % event.get("name"))
                unprocessed.append(event)
                continue
        else:
            prev_evt = unprocessed.pop()
            if event_name(event) == event_name(prev_evt):
                if start_time:
                    record = event_record(start_time, prev_evt, event)
                    records.append(
                        format_record(
                            "Finished stage: (%n) %d seconds", record
                        )
                        + "\n"
                    )
                    total_time += record.get("delta") or 0.0
            else:
                # not a match, put it back
                unprocessed.append(prev_evt)

    records.append(total_time_record(total_time))
    boot_records.append(records)
    return boot_records


def show_events(events: List[Event], print_format: str) -> List[List[str]]:
    """
    A passthrough method that makes it easier to call generate_records()

    :param events: JSONs from dump that represents events taken from logs
    :param print_format: formatting to represent event, time stamp,
    and time taken by the event in one line

    :return: boot records ordered chronologically
    """
    return generate_records(events, print_format=print_format)


def load_events_infile(infile: IO) -> Tuple[Optional[Any], str]:
    """
    Takes in a log file, read it, and convert to json.

    :param infile: The Log file to be read

    :return: json version of logfile, raw file
    """
    data = infile.read()
    if not data.strip():
        sys.stderr.write("Empty file %s\n" % infile.name)
        sys.exit(1)
    try:
        return json.loads(data), data
    except ValueError:
        return None, data



# This file is part of cloud-init. See LICENSE file for license information.

from unittest import mock

import pytest

from cloudinit import distros, util


class TestAlpineBusyboxUserGroup:
    @mock.patch("cloudinit.distros.alpine.subp.subp")
    @mock.patch("cloudinit.distros.subp.which", return_value=False)
    def test_busybox_add_group(self, m_which, m_subp):
        distro = distros.fetch("alpine")("alpine", {}, None)

        group = "mygroup"

        distro.create_group(group)

        m_subp.assert_called_with(["addgroup", group])

    @pytest.mark.usefixtures("fake_filesystem")
    @mock.patch("cloudinit.distros.alpine.subp.subp")
    @mock.patch("cloudinit.distros.subp.which", return_value=False)
    def test_busybox_add_user(self, m_which, m_subp, tmpdir):
        distro = distros.fetch("alpine")("alpine", {}, None)

        shadow_file = tmpdir.join("/etc/shadow")
        shadow_file.dirpath().mkdir()

        user = "me2"

        # Need to place entry for user in /etc/shadow as
        # "adduser" is stubbed and so will not create it.
        root_entry = "root::19848:0:::::"
        shadow_file.write(
            root_entry + "\n" + user + ":!:19848:0:99999:7:::" + "\n"
        )

        distro.shadow_fn = shadow_file

        distro.add_user(user, lock_passwd=True)

        m_subp.assert_called_with(["adduser", "-D", user])

        contents = util.load_text_file(shadow_file)
        expected = root_entry + "\n" + user + ":!:19848::::::" + "\n"

        assert contents == expected


class TestAlpineShadowUserGroup:
    distro = distros.fetch("alpine")("alpine", {}, None)

    @mock.patch("cloudinit.distros.alpine.subp.subp")
    @mock.patch(
        "cloudinit.distros.subp.which", return_value=("/usr/sbin/groupadd")
    )
    def test_shadow_add_group(self, m_which, m_subp):
        group = "mygroup"

        self.distro.create_group(group)

        m_subp.assert_called_with(["groupadd", group])

    @mock.patch("cloudinit.distros.alpine.subp.subp")
    @mock.patch(
        "cloudinit.distros.subp.which", return_value=("/usr/sbin/useradd")
    )
    def test_shadow_add_user(self, m_which, m_subp):
        user = "me2"

        self.distro.add_user(user)

        m_subp.assert_called_with(
            ["useradd", user, "-m"], logstring=["useradd", user, "-m"]
        )



# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init

import copy
import os
import random
import shutil
import string
import time
import unittest
from contextlib import contextmanager
from unittest import mock
from unittest.util import strclass
from urllib.parse import urlsplit, urlunsplit

import pytest
import responses

from cloudinit import distros, helpers, settings, util
from cloudinit.helpers import Paths
from cloudinit.templater import JINJA_AVAILABLE
from tests.helpers import cloud_init_project_dir

try:
    import apt_pkg  # type: ignore # noqa: F401

    HAS_APT_PKG = True
except ImportError:
    HAS_APT_PKG = False


# Used by tests to verify the error message when a jsonschema structure
# is empty but should not be.
# Version 4.20.0 of jsonschema changed the error messages for empty structures.
SCHEMA_EMPTY_ERROR = (
    "(is too short|should be non-empty|does not have enough properties)"
)

example_netdev = {
    "eth0": {
        "hwaddr": "00:16:3e:16:db:54",
        "ipv4": [
            {
                "bcast": "10.85.130.255",
                "ip": "10.85.130.116",
                "mask": "255.255.255.0",
                "scope": "global",
            }
        ],
        "ipv6": [
            {
                "ip": "fd42:baa2:3dd:17a:216:3eff:fe16:db54/64",
                "scope6": "global",
            },
            {"ip": "fe80::216:3eff:fe16:db54/64", "scope6": "link"},
        ],
        "up": True,
    },
    "lo": {
        "hwaddr": "",
        "ipv4": [
            {
                "bcast": "",
                "ip": "127.0.0.1",
                "mask": "255.0.0.0",
                "scope": "host",
            }
        ],
        "ipv6": [{"ip": "::1/128", "scope6": "host"}],
        "up": True,
    },
}


# Makes the old path start
# with new base instead of whatever
# it previously had
def rebase_path(old_path, new_base):
    if old_path.startswith(new_base):
        # Already handled...
        return old_path
    # Retarget the base of that path
    # to the new base instead of the
    # old one...
    path = os.path.join(new_base, old_path.lstrip("/"))
    path = os.path.abspath(path)
    return path


# Can work on anything that takes a path as arguments
def retarget_many_wrapper(new_base, am, old_func):
    def wrapper(*args, **kwds):
        n_args = list(args)
        nam = am
        if am == -1:
            nam = len(n_args)
        for i in range(nam):
            path = args[i]
            # patchOS() wraps various os and os.path functions, however in
            # Python 3 some of these now accept file-descriptors (integers).
            # That breaks rebase_path() so in lieu of a better solution, just
            # don't rebase if we get a fd.
            if isinstance(path, str):
                n_args[i] = rebase_path(path, new_base)
        return old_func(*n_args, **kwds)

    return wrapper


def random_string(length=8):
    """return a random lowercase string with default length of 8"""
    return "".join(
        random.choice(string.ascii_lowercase) for _ in range(length)
    )


# Note: The use of this class and unittests.TestCase is discouraged. Use pytest
# instead. See development docs on testing.
class TestCase(unittest.TestCase):
    def reset_global_state(self):
        """Reset any global state to its original settings.

        cloudinit caches some values in cloudinit.util.  Unit tests that
        involved those cached paths were then subject to failure if the order
        of invocation changed (LP: #1703697).

        This function resets any of these global state variables to their
        initial state.

        In the future this should really be done with some registry that
        can then be cleaned in a more obvious way.
        """
        util._DNS_REDIRECT_IP = None

    def setUp(self):
        super(TestCase, self).setUp()
        self.reset_global_state()

    def shortDescription(self):
        return strclass(self.__class__) + "." + self._testMethodName

    def add_patch(self, target, attr, *args, **kwargs):
        """Patches specified target object and sets it as attr on test
        instance also schedules cleanup"""
        if "autospec" not in kwargs:
            kwargs["autospec"] = True
        m = mock.patch(target, *args, **kwargs)
        p = m.start()
        self.addCleanup(m.stop)
        setattr(self, attr, p)


def replicate_test_root(example_root, target_root):
    real_root = resourceLocation()
    real_root = os.path.join(real_root, "roots", example_root)
    for dir_path, _dirnames, filenames in os.walk(real_root):
        real_path = dir_path
        make_path = rebase_path(real_path[len(real_root) :], target_root)
        util.ensure_dir(make_path)
        for f in filenames:
            real_path = os.path.abspath(os.path.join(real_path, f))
            make_path = os.path.abspath(os.path.join(make_path, f))
            shutil.copy(real_path, make_path)


def responses_assert_call_count(url: str, count: int) -> bool:
    """Focal and older have a version of responses which does
    not carry this attribute. This can be removed when focal
    is no longer supported.
    """
    if hasattr(responses, "assert_call_count"):
        return responses.assert_call_count(url, count)

    def _ensure_url_default_path(url):
        if isinstance(url, str):
            url_parts = list(urlsplit(url))
            if url_parts[2] == "":
                url_parts[2] = "/"
                url = urlunsplit(url_parts)
        return url

    call_count = len(
        [
            1
            for call in responses.calls
            if call.request.url == _ensure_url_default_path(url)
        ]
    )
    if call_count == count:
        return True
    else:
        raise AssertionError(
            f"Expected URL '{url}' to be called {count} times. "
            f"Called {call_count} times."
        )


def get_mock_paths(temp_dir):
    class MockPaths(Paths):
        def __init__(self, path_cfgs: dict, ds=None):
            super().__init__(path_cfgs=path_cfgs, ds=ds)

            self.cloud_dir: str = path_cfgs.get(
                "cloud_dir", f"{temp_dir}/var/lib/cloud"
            )
            self.run_dir: str = path_cfgs.get(
                "run_dir", f"{temp_dir}/run/cloud/"
            )
            self.template_dir: str = path_cfgs.get(
                "templates_dir", f"{temp_dir}/etc/cloud/templates/"
            )

    return MockPaths


def populate_dir(path, files):
    if not os.path.exists(path):
        os.makedirs(path)
    ret = []
    for name, content in files.items():
        p = os.path.sep.join([path, name])
        util.ensure_dir(os.path.dirname(p))
        with open(p, "wb") as fp:
            if isinstance(content, bytes):
                fp.write(content)
            else:
                fp.write(content.encode("utf-8"))
            fp.close()
        ret.append(p)

    return ret


def populate_dir_with_ts(path, data):
    """data is {'file': ('contents', mtime)}.  mtime relative to now."""
    populate_dir(path, dict((k, v[0]) for k, v in data.items()))
    btime = time.time()
    for fpath, (_contents, mtime) in data.items():
        ts = btime + mtime if mtime else btime
        os.utime(os.path.sep.join((path, fpath)), (ts, ts))


def dir2dict(startdir, prefix=None, filter=None):
    flist = {}
    if prefix is None:
        prefix = startdir
    for root, _dirs, files in os.walk(startdir):
        for fname in files:
            fpath = os.path.join(root, fname)
            key = fpath[len(prefix) :]
            if filter is not None and not filter(fpath):
                continue
            flist[key] = util.load_text_file(fpath)
    return flist


def wrap_and_call(prefix, mocks, func, *args, **kwargs):
    """
    call func(args, **kwargs) with mocks applied, then unapplies mocks
    nicer to read than repeating dectorators on each function

    prefix: prefix for mock names (e.g. 'cloudinit.stages.util') or None
    mocks: dictionary of names (under 'prefix') to mock and either
        a return value or a dictionary to pass to the mock.patch call
    func: function to call with mocks applied
    *args,**kwargs: arguments for 'func'

    return_value: return from 'func'
    """
    delim = "."
    if prefix is None:
        prefix = ""
    prefix = prefix.rstrip(delim)
    unwraps = []
    for fname, kw in mocks.items():
        if prefix:
            fname = delim.join((prefix, fname))
        if not isinstance(kw, dict):
            kw = {"return_value": kw}
        p = mock.patch(fname, **kw)
        p.start()
        unwraps.append(p)
    try:
        return func(*args, **kwargs)
    finally:
        for p in unwraps:
            p.stop()


def resourceLocation(subname=None):
    path = cloud_init_project_dir("tests/data")
    if not subname:
        return path
    return os.path.join(path, subname)


def readResource(name, mode="r"):
    with open(resourceLocation(name), mode) as fh:
        return fh.read()


def skipIfAptPkg():
    return pytest.mark.skipif(
        HAS_APT_PKG, reason="No python-apt dependency present."
    )


try:
    import importlib.metadata

    import jsonschema

    assert jsonschema  # avoid pyflakes error F401: import unused
    _jsonschema_version = tuple(
        int(part)
        for part in importlib.metadata.metadata("jsonschema")
        .get("Version", "")
        .split(".")
    )
    _missing_jsonschema_dep = False
except ImportError:
    _missing_jsonschema_dep = True
    _jsonschema_version = (0, 0, 0)


def skipUnlessJsonSchemaVersionGreaterThan(version=(0, 0, 0)):
    return pytest.mark.skipif(
        _jsonschema_version <= version,
        reason=(
            f"python3-jsonschema {_jsonschema_version} not greater than"
            f" {version}"
        ),
    )


def skipUnlessJsonSchema():
    return pytest.mark.skipif(
        _missing_jsonschema_dep,
        reason="No python-jsonschema dependency present.",
    )


def skipUnlessJinja():
    return pytest.mark.skipif(
        not JINJA_AVAILABLE, reason="No jinja dependency present."
    )


@skipUnlessJinja()
def skipUnlessJinjaVersionGreaterThan(version=(0, 0, 0)):
    import jinja2

    return pytest.mark.skipif(
        tuple(map(int, jinja2.__version__.split("."))) < version,
        reason=f"jinj2 version is less than {version}",
    )


def skipIfJinja():
    return pytest.mark.skipif(
        JINJA_AVAILABLE, reason="Jinja dependency present."
    )


# older versions of mock do not have the useful 'assert_not_called'
if not hasattr(mock.Mock, "assert_not_called"):

    def __mock_assert_not_called(mmock):
        if mmock.call_count != 0:
            msg = (
                "[citest] Expected '%s' to not have been called. "
                "Called %s times."
                % (mmock._mock_name or "mock", mmock.call_count)
            )
            raise AssertionError(msg)

    mock.Mock.assert_not_called = __mock_assert_not_called  # type: ignore


@contextmanager
def does_not_raise():
    """Context manager to parametrize tests raising and not raising exceptions

    Note: In python-3.7+, this can be substituted by contextlib.nullcontext
    More info:
    https://docs.pytest.org/en/6.2.x/example/parametrize.html?highlight=does_not_raise#parametrizing-conditional-raising

    Example:
    --------
    >>> @pytest.mark.parametrize(
    >>>     "example_input,expectation",
    >>>     [
    >>>         (1, does_not_raise()),
    >>>         (0, pytest.raises(ZeroDivisionError)),
    >>>     ],
    >>> )
    >>> def test_division(example_input, expectation):
    >>>     with expectation:
    >>>         assert (0 / example_input) is not None

    """
    yield


def get_distro(dname, system_info=None, /, renderers=None, activators=None):
    """Return a Distro class of distro 'dname'.

    system_info has the format of CFG_BUILTIN['system_info'].

    Example: get_distro("debian")
    """
    if system_info is None:
        system_info = copy.deepcopy(settings.CFG_BUILTIN["system_info"])
    system_info["distro"] = dname
    if renderers:
        system_info["network"]["renderers"] = renderers
    if activators:
        system_info["network"]["activators"] = activators
    paths = helpers.Paths(system_info["paths"])
    distro_cls = distros.fetch(dname)
    return distro_cls(dname, system_info, paths)


def assert_count_equal(a, b):
    """
    Equivalent to unittests.TestCase.assertCountEqual.

    https://docs.python.org/3/library/unittest.html#unittest.TestCase.assertCountEqual
    """
    case = unittest.TestCase()
    case.assertCountEqual(a, b)



# This file is part of cloud-init. See LICENSE file for license information.

import os
import stat
from collections import namedtuple
from unittest import mock

import pytest

import cloudinit.settings
from cloudinit.cmd import clean
from cloudinit.config import cc_mounts
from cloudinit.distros import Distro
from cloudinit.sources import DataSource
from cloudinit.stages import Init
from cloudinit.util import del_file, ensure_dir, sym_link, write_file
from tests.unittests.helpers import wrap_and_call

MyPaths = namedtuple("MyPaths", "cloud_dir")
CleanPaths = namedtuple(
    "CleanPaths",
    ["tmpdir", "cloud_dir", "clean_dir", "log", "output_log"],
)


@pytest.fixture(scope="function")
def clean_paths(tmpdir):
    return CleanPaths(
        tmpdir=tmpdir,
        cloud_dir=tmpdir.join("varlibcloud"),
        clean_dir=tmpdir.join("clean.d"),
        log=tmpdir.join("cloud-init.log"),
        output_log=tmpdir.join("cloud-init-output.log"),
    )


@pytest.fixture(scope="function")
def init_class(clean_paths):
    init = mock.Mock(spec=Init)
    init.paths = MyPaths(cloud_dir=f"{clean_paths.cloud_dir}/")
    init.distro.shutdown_command = Distro.shutdown_command
    init.cfg = {
        "def_log_file": clean_paths.log,
        "output": {"all": f"|tee -a {clean_paths.output_log}"},
    }
    return init


class TestClean:
    def test_remove_artifacts_removes_logs(self, clean_paths, init_class):
        """remove_artifacts removes logs when remove_logs is True."""
        clean_paths.log.write("cloud-init-log")
        clean_paths.output_log.write("cloud-init-output-log")

        assert (
            os.path.exists(clean_paths.cloud_dir) is False
        ), "Unexpected cloud_dir"
        retcode = clean.remove_artifacts(
            init=init_class,
            remove_logs=True,
        )
        assert (
            clean_paths.log.exists() is False
        ), f"Unexpected file {clean_paths.log}"
        assert (
            clean_paths.output_log.exists() is False
        ), f"Unexpected file {clean_paths.output_log}"
        assert 0 == retcode

    def test_remove_net_conf(self, clean_paths, init_class):
        """remove_config removes config files when True."""

        TEST_GEN_NET_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_NET_CONFIG_FILES
        ]
        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected {conf_path}"
            ensure_dir(os.path.dirname(conf_path))
            if "*" in conf_path.strpath:
                # Expand glob path to interface-specific eth0 filename
                eth0_path = conf_path.strpath.replace("*", "eth0")
                write_file(
                    eth0_path, f"#generated by cloud-init test {eth0_path}"
                )
            else:
                conf_path.write(f"#generated by cloud-init test {conf_path}")

        with mock.patch(
            "cloudinit.cmd.clean.GEN_NET_CONFIG_FILES",
            [f.strpath for f in TEST_GEN_NET_CONFIG_FILES],
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_config=["network"],
            )

        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            if "*" in conf_path.strpath:
                # Expand glob path to interface-specific eth0 filename
                assert (
                    os.path.exists(conf_path.strpath.replace("*", "eth0"))
                    is False
                ), f"Unexpected file {conf_path.strpath.replace('*', 'eth0')}"
            else:
                assert (
                    conf_path.exists() is False
                ), f"Unexpected file {conf_path}"
        assert 0 == retcode

    def test_remove_ssh_conf(self, clean_paths, init_class):
        """remove_config removes config files when True."""

        TEST_GEN_SSH_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_SSH_CONFIG_FILES
        ]
        for conf_path in TEST_GEN_SSH_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected {conf_path}"
            ensure_dir(os.path.dirname(conf_path))
            conf_path.write(f"#generated by cloud-init\ntouch {conf_path}\n")

        with mock.patch(
            "cloudinit.cmd.clean.GEN_SSH_CONFIG_FILES",
            TEST_GEN_SSH_CONFIG_FILES,
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_config=["ssh_config"],
            )

        for conf_path in TEST_GEN_SSH_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected file {conf_path}"
        assert 0 == retcode

    def test_remove_all_conf(self, clean_paths, init_class):
        """remove_config removes config files when True."""

        TEST_GEN_NET_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_NET_CONFIG_FILES
        ]
        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected {conf_path}"
            ensure_dir(os.path.dirname(conf_path))
            conf_path.write(f"#generated by cloud-init\ntouch {conf_path}\n")

        TEST_GEN_SSH_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_SSH_CONFIG_FILES
        ]
        for conf_path in TEST_GEN_SSH_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected {conf_path}"
            ensure_dir(os.path.dirname(conf_path))
            conf_path.write(f"#generated by cloud-init\ntouch {conf_path}\n")

        with mock.patch(
            "cloudinit.cmd.clean.GEN_NET_CONFIG_FILES",
            [f.strpath for f in TEST_GEN_NET_CONFIG_FILES],
        ), mock.patch(
            "cloudinit.cmd.clean.GEN_SSH_CONFIG_FILES",
            TEST_GEN_SSH_CONFIG_FILES,
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_config=["all"],
            )

        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            assert conf_path.exists() is False, f"file {conf_path} exists!"
        for conf_path in TEST_GEN_SSH_CONFIG_FILES:
            assert conf_path.exists() is False, f"file {conf_path} exists!"
        assert 0 == retcode

    def test_keep_net_conf(self, clean_paths, init_class):
        """remove_config removes config files when True."""

        TEST_GEN_NET_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_NET_CONFIG_FILES
        ]
        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected {conf_path}"
            ensure_dir(os.path.dirname(conf_path))
            conf_path.write(f"#generated by cloud-init\ntouch {conf_path}\n")

        with mock.patch(
            "cloudinit.cmd.clean.GEN_NET_CONFIG_FILES",
            TEST_GEN_NET_CONFIG_FILES,
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_config=[],
            )

        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            assert conf_path.exists() is True, f"file {conf_path} removed!"
        assert 0 == retcode

    @pytest.mark.usefixtures("fake_filesystem")
    def test_clean_fstab(self, clean_paths, init_class):
        """remove_config removed added entries in fstab when
        `cloud-init clean -c fstab` is used.
        """
        fstab_original_content = (
            "UUID=abc123 / ext4 defaults 0 0\n"
            "/workspace	/mnt	"
            "auto	defaults,nofail,x-systemd.after="
            "cloud-init.service,_netdev,comment=cloudconfig	0	2\n"
        )
        fstab_expected_content = "UUID=abc123 / ext4 defaults 0 0\n"

        etc_path = "/etc"
        if not os.path.exists(etc_path):
            os.makedirs(etc_path)
        fstab_path = cc_mounts.FSTAB_PATH
        with open(fstab_path, "w") as fd:
            fd.write(fstab_original_content)

        clean.remove_artifacts(
            init_class,
            remove_logs=False,
            remove_config=["fstab"],
        )

        with open(fstab_path, "r") as fd:
            fstab_new_content = fd.read()
            assert fstab_expected_content == fstab_new_content

    @pytest.mark.usefixtures("fake_filesystem")
    def test_clean_fstab_for_all(self, clean_paths, init_class):
        """remove_config removed added entries in fstab when
        `cloud-init clean -c all` is used.
        """
        fstab_original_content = (
            "UUID=abc123 / ext4 defaults 0 0\n"
            "/workspace	/mnt	"
            "auto	defaults,nofail,x-systemd.after="
            "cloud-init.service,_netdev,comment=cloudconfig	0	2\n"
        )
        fstab_expected_content = "UUID=abc123 / ext4 defaults 0 0\n"
        TEST_GEN_SSH_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_SSH_CONFIG_FILES
        ]

        etc_path = "/etc"
        if not os.path.exists(etc_path):
            os.makedirs(etc_path)
        fstab_path = cc_mounts.FSTAB_PATH
        with open(fstab_path, "w") as fd:
            fd.write(fstab_original_content)

        with mock.patch(
            "cloudinit.cmd.clean.GEN_SSH_CONFIG_FILES",
            TEST_GEN_SSH_CONFIG_FILES,
        ):

            clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_config=["all"],
            )

        with open(fstab_path, "r") as fd:
            fstab_new_content = fd.read()
            assert fstab_expected_content == fstab_new_content

    def test_clean_datasource_conf_without_cache(
        self, clean_paths, init_class
    ):
        """remove_config does not remove datasource files when cache
        is not present.
        """
        ds_conf = clean_paths.tmpdir.join("/var/run/ds")
        assert ds_conf.exists() is False, f"Unexpected {ds_conf}"
        ensure_dir(os.path.dirname(ds_conf))
        ds_conf.write("#generated by generic DataSource\nfoobar\n")
        assert ds_conf.exists() is True, "{ds_conf} not written!"

        assert (
            clean_paths.cloud_dir.exists() is False
        ), "unexpected cloud_dir present!"

        def ds_clean():
            print("deleting {ds_conf}")
            del_file(ds_conf)

        ds = mock.Mock(spec=DataSource)
        ds.clean = ds_clean

        def ds_fetch():
            return ds

        init_class.fetch = ds_fetch
        retcode = clean.remove_artifacts(
            init_class,
            remove_logs=False,
            remove_config=["datasource"],
        )
        assert ds_conf.exists() is True, f"file {ds_conf} was removed!"
        assert 0 == retcode

    def test_clean_datasource_conf_with_cache(self, clean_paths, init_class):
        """remove_config removes datasource files when cache
        is present.
        """
        ensure_dir(clean_paths.cloud_dir)
        ds_conf = clean_paths.tmpdir.join("/var/run/ds")
        assert ds_conf.exists() is False, f"Unexpected {ds_conf}"
        ensure_dir(os.path.dirname(ds_conf))
        ds_conf.write("#generated by generic DataSource\nfoobar\n")
        assert ds_conf.exists() is True, "{ds_conf} not written!"

        cache = clean_paths.cloud_dir.join("instance")
        assert cache.exists() is False, f"unexpected {cache} present!"
        ensure_dir(cache)
        assert cache.exists() is True, "{cache} not created!"

        def ds_clean():
            print("deleting {ds_conf}")
            del_file(ds_conf)

        ds = mock.Mock(spec=DataSource)
        ds.clean = ds_clean

        def ds_fetch():
            return ds

        init_class.fetch = ds_fetch
        with mock.patch(
            "cloudinit.cmd.clean.settings.CLEAN_RUNPARTS_DIR", os.devnull
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_config=["datasource"],
            )
        assert ds_conf.exists() is False, f"Unexpected file {ds_conf}"
        assert 0 == retcode

    @pytest.mark.allow_all_subp
    def test_remove_artifacts_runparts_clean_d(self, clean_paths, init_class):
        """remove_artifacts performs runparts on CLEAN_RUNPARTS_DIR"""
        ensure_dir(clean_paths.cloud_dir)
        artifact_file = clean_paths.tmpdir.join("didit")
        ensure_dir(clean_paths.clean_dir)
        assert artifact_file.exists() is False, f"Unexpected {artifact_file}"
        clean_script = clean_paths.clean_dir.join("1.sh")
        clean_script.write(f"#!/bin/sh\ntouch {artifact_file}\n")
        clean_script.chmod(mode=0o755)
        with mock.patch.object(
            cloudinit.settings, "CLEAN_RUNPARTS_DIR", clean_paths.clean_dir
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
            )
        assert (
            artifact_file.exists() is True
        ), f"Missing expected {artifact_file}"
        assert 0 == retcode

    def test_remove_artifacts_preserves_logs(self, clean_paths, init_class):
        """remove_artifacts leaves logs when remove_logs is False."""
        clean_paths.log.write("cloud-init-log")
        clean_paths.output_log.write("cloud-init-output-log")

        retcode = clean.remove_artifacts(
            init_class,
            remove_logs=False,
        )
        assert 0 == retcode
        assert (
            clean_paths.log.exists() is True
        ), f"Missing expected file {clean_paths.log}"
        assert (
            clean_paths.output_log.exists()
        ), f"Missing expected file {clean_paths.output_log}"

    def test_remove_artifacts_removes_unlinks_symlinks(
        self, clean_paths, init_class
    ):
        """remove_artifacts cleans artifacts dir unlinking any symlinks."""
        dir1 = clean_paths.cloud_dir.join("dir1")
        ensure_dir(dir1)
        symlink = clean_paths.cloud_dir.join("mylink")
        sym_link(dir1.strpath, symlink.strpath)

        with mock.patch.object(
            cloudinit.settings, "CLEAN_RUNPARTS_DIR", clean_paths.clean_dir
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
            )
        assert 0 == retcode
        for path in (dir1, symlink):
            assert path.exists() is False, f"Unexpected {path} found"

    def test_remove_artifacts_removes_artifacts_skipping_seed(
        self, clean_paths, init_class
    ):
        """remove_artifacts cleans artifacts dir with exception of seed dir."""
        dirs = [
            clean_paths.cloud_dir,
            clean_paths.cloud_dir.join("seed"),
            clean_paths.cloud_dir.join("dir1"),
            clean_paths.cloud_dir.join("dir2"),
        ]
        for _dir in dirs:
            ensure_dir(_dir)

        with mock.patch.object(
            cloudinit.settings, "CLEAN_RUNPARTS_DIR", clean_paths.clean_dir
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
            )
        assert 0 == retcode
        for expected_dir in dirs[:2]:
            assert expected_dir.exists() is True, f"Missing {expected_dir}"
        for deleted_dir in dirs[2:]:
            assert deleted_dir.exists() is False, f"Unexpected {deleted_dir}"

    def test_remove_artifacts_removes_artifacts_removes_seed(
        self, clean_paths, init_class
    ):
        """remove_artifacts removes seed dir when remove_seed is True."""
        dirs = [
            clean_paths.cloud_dir,
            clean_paths.cloud_dir.join("seed"),
            clean_paths.cloud_dir.join("dir1"),
            clean_paths.cloud_dir.join("dir2"),
        ]
        for _dir in dirs:
            ensure_dir(_dir)

        with mock.patch.object(
            cloudinit.settings, "CLEAN_RUNPARTS_DIR", clean_paths.clean_dir
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_seed=True,
            )
        assert 0 == retcode
        assert (
            clean_paths.cloud_dir.exists() is True
        ), f"Missing dir {clean_paths.cloud_dir}"
        for deleted_dir in dirs[1:]:
            assert (
                deleted_dir.exists() is False
            ), f"Unexpected {deleted_dir} dir"

    def test_remove_artifacts_returns_one_on_errors(
        self, clean_paths, init_class, capsys
    ):
        """remove_artifacts returns non-zero on failure and prints an error."""
        ensure_dir(clean_paths.cloud_dir)
        ensure_dir(clean_paths.cloud_dir.join("dir1"))

        with mock.patch(
            "cloudinit.cmd.clean.del_dir", side_effect=OSError("oops")
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
            )
        assert 1 == retcode
        _out, err = capsys.readouterr()
        assert (
            f"Error:\nCould not remove {clean_paths.cloud_dir}/dir1: oops\n"
            == err
        )

    def test_handle_clean_args_reboots(self, init_class):
        """handle_clean_args_reboots when reboot arg is provided."""

        called_cmds = []

        def fake_subp(cmd, capture):
            called_cmds.append(cmd)
            return "", ""

        myargs = namedtuple(
            "myargs", "remove_logs remove_seed remove_config reboot machine_id"
        )
        cmdargs = myargs(
            remove_logs=False,
            remove_seed=False,
            remove_config=[],
            reboot=True,
            machine_id=False,
        )
        retcode = wrap_and_call(
            "cloudinit.cmd.clean",
            {
                "subp": {"side_effect": fake_subp},
                "Init": {"return_value": init_class},
            },
            clean.handle_clean_args,
            name="does not matter",
            args=cmdargs,
        )
        assert 0 == retcode
        assert [["shutdown", "-r", "now"]] == called_cmds

    @pytest.mark.parametrize(
        "machine_id,systemd_val",
        (
            pytest.param(True, True, id="machine_id_on_systemd_uninitialized"),
            pytest.param(
                True, False, id="machine_id_non_systemd_removes_file"
            ),
            pytest.param(False, False, id="no_machine_id_param_file_remains"),
        ),
    )
    @mock.patch("cloudinit.cmd.clean.uses_systemd")
    def test_handle_clean_args_removed_machine_id(
        self, uses_systemd, machine_id, systemd_val, clean_paths, init_class
    ):
        """handle_clean_args removes /etc/machine-id when arg is True."""
        uses_systemd.return_value = systemd_val
        myargs = namedtuple(
            "myargs", "remove_logs remove_seed remove_config reboot machine_id"
        )
        cmdargs = myargs(
            remove_logs=False,
            remove_seed=False,
            remove_config=[],
            reboot=False,
            machine_id=machine_id,
        )
        machine_id_path = clean_paths.tmpdir.join("machine-id")
        machine_id_path.write("SOME-AMAZN-MACHINE-ID")
        with mock.patch.object(
            cloudinit.settings, "CLEAN_RUNPARTS_DIR", clean_paths.clean_dir
        ):
            with mock.patch.object(
                cloudinit.cmd.clean, "ETC_MACHINE_ID", machine_id_path.strpath
            ):
                with mock.patch(
                    "cloudinit.cmd.clean.Init", return_value=init_class
                ):
                    assert 0 == clean.handle_clean_args(
                        name="does not matter",
                        args=cmdargs,
                    )
        if systemd_val:
            if machine_id:
                assert "uninitialized\n" == machine_id_path.read()
            else:
                assert "SOME-AMAZN-MACHINE-ID" == machine_id_path.read()
        else:
            assert machine_id_path.exists() is bool(not machine_id)

    def test_status_main(self, clean_paths, init_class):
        """clean.main can be run as a standalone script."""
        clean_paths.log.write("cloud-init-log")
        with pytest.raises(SystemExit) as context_manager:
            wrap_and_call(
                "cloudinit.cmd.clean",
                {
                    "Init": {"return_value": init_class},
                    "sys.argv": {"new": ["clean", "--logs"]},
                },
                clean.main,
            )
        assert 0 == context_manager.value.code
        assert (
            clean_paths.log.exists() is False
        ), f"Unexpected log {clean_paths.log}"

    @pytest.mark.parametrize("device_type", (stat.S_IFCHR, stat.S_IFBLK))
    @pytest.mark.usefixtures("fake_filesystem")
    def test_remove_artifacts_preserves_device_logs(
        self, device_type, clean_paths, init_class
    ):
        """remove_artifacts does not remove log files that are device files."""
        original_stat = os.stat

        def mock_stat(path):
            if path == clean_paths.log.strpath:
                st_result = mock.Mock()
                st_result.st_mode = device_type | 0o666
                return st_result
            return original_stat(path)

        # Create a regular file and mock its stat
        clean_paths.log.write("device-file")
        clean_paths.output_log.write("cloud-init-output-log")
        with mock.patch("cloudinit.cmd.clean.os.stat", side_effect=mock_stat):
            retcode = clean.remove_artifacts(
                init=init_class,
                remove_logs=True,
            )
            assert 0 == retcode
            assert (
                clean_paths.log.exists() is True
            ), f"Device file {clean_paths.log} should not be removed"
            assert (
                clean_paths.output_log.exists() is False
            ), f"Regular log {clean_paths.output_log} should be removed"



# This file is part of cloud-init. See LICENSE file for license information.

import cloudinit.util
from tests.unittests.helpers import get_distro, mock

M_PATH = "cloudinit.distros."


class TestDragonFlyBSD:
    @mock.patch(M_PATH + "subp.subp")
    def test_add_user(self, m_subp):
        distro = get_distro("dragonflybsd")
        assert True is distro.add_user("me2", uid=1234, default=False)
        assert [
            mock.call(
                [
                    "pw",
                    "useradd",
                    "-n",
                    "me2",
                    "-u",
                    "1234",
                    "-d/home/me2",
                    "-m",
                ],
                logstring=["pw", "useradd", "-n", "me2", "-d/home/me2", "-m"],
            )
        ] == m_subp.call_args_list

    def test_unlock_passwd(self, caplog):
        distro = get_distro("dragonflybsd")
        distro.unlock_passwd("me2")
        assert (
            "Dragonfly BSD/FreeBSD password lock is not reversible, "
            "ignoring unlock for user me2" in caplog.text
        )


def test_find_dragonflybsd_part():
    assert cloudinit.util.find_freebsd_part("/dev/vbd0s3") == "vbd0s3"


@mock.patch("cloudinit.util.is_DragonFlyBSD")
@mock.patch("cloudinit.subp.subp")
def test_parse_mount(mock_subp, m_is_DragonFlyBSD):
    mount_out = """
vbd0s3 on / (hammer2, local)
devfs on /dev (devfs, nosymfollow, local)
/dev/vbd0s0a on /boot (ufs, local)
procfs on /proc (procfs, local)
tmpfs on /var/run/shm (tmpfs, local)
"""

    mock_subp.return_value = (mount_out, "")
    m_is_DragonFlyBSD.return_value = True
    assert cloudinit.util.parse_mount("/") == ("vbd0s3", "hammer2", "/")



# Copyright (C) 2021-2025 VMware by Broadcom.
#
# Author: Shreenidhi Shedi <yesshedi@gmail.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import logging
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Tuple

from cloudinit import subp, util
from cloudinit.net import renderer, should_add_gateway_onlink_flag
from cloudinit.net.network_state import NetworkState

LOG = logging.getLogger(__name__)


def normalize(data: Dict[str, List[Any]]) -> Dict[str, List[Any]]:
    """
    Normalize a dictionary of lists.
    - Assumes top-level keys map to lists.
    - Each list and any nested dicts/lists will be recursively normalized.
    """
    normalized = {}
    for key, value in data.items():
        normalized[key] = _normalize_value(value)
    return normalized


def _normalize_value(data: Any) -> Any:
    """
    Recursively normalize a dictionary or list:
    - Dicts: keys sorted, values normalized
    - Lists: items normalized and sorted (if comparable)
    """
    if isinstance(data, dict):
        normalized = {}
        for key in sorted(data):
            normalized[key] = _normalize_value(data[key])
        return normalized
    elif isinstance(data, list):
        normalized_items = []
        for item in data:
            if isinstance(item, (dict, list)):
                normalized_item = _normalize_value(item)
            else:
                normalized_item = item
            normalized_items.append(normalized_item)
        try:
            return sorted(normalized_items)
        except TypeError:
            return normalized_items

    return data


class CfgParser:
    def __init__(self):
        self.conf_dict = {
            "Match": [],
            "Link": [],
            "Network": [],
            "DHCPv4": [],
            "DHCPv6": [],
            "Address": [],
            "Route": {},
            "NetDev": [],
            "VLAN": [],
            "Bond": [],
            "Bridge": [],
            "RoutingPolicyRule": {},
        }

    def update_section(self, sec, key, val):
        for k in self.conf_dict.keys():
            if k == sec:
                self.conf_dict[k].append(f"{key}={val}")
                self.conf_dict[k] = list(dict.fromkeys(self.conf_dict[k]))
                self.conf_dict[k].sort()

    def update_route_section(self, sec, rid, key, val):
        """
        For each route section we use rid as a key, this allows us to isolate
        this route from others on subsequent calls.
        """
        for k in self.conf_dict.keys():
            if k == sec:
                if rid not in self.conf_dict[k]:
                    self.conf_dict[k][rid] = []
                self.conf_dict[k][rid].append(f"{key}={val}")
                # remove duplicates from list
                self.conf_dict[k][rid] = list(
                    dict.fromkeys(self.conf_dict[k][rid])
                )
                self.conf_dict[k][rid].sort()

    def get_final_conf(self):
        contents = ""

        self.conf_dict = normalize(self.conf_dict)

        for k, v in sorted(self.conf_dict.items()):
            if not v:
                continue
            if k == "Address":
                for e in sorted(v):
                    contents += f"[{k}]\n{e}\n\n"
            elif k in ["Route", "RoutingPolicyRule"]:
                for n in sorted(v):
                    contents += f"[{k}]\n"
                    for e in sorted(v[n]):
                        contents += f"{e}\n"
                    contents += "\n"
            else:
                contents += f"[{k}]\n"
                for e in sorted(v):
                    contents += f"{e}\n"
                contents += "\n"

        return contents


class Renderer(renderer.Renderer):
    """
    Renders network information in /etc/systemd/network

    This Renderer is currently experimental and doesn't support all the
    use cases supported by the other renderers yet.
    """

    def __init__(self, config=None):
        if not config:
            config = {}
        self.resolve_conf_fn = config.get(
            "resolve_conf_fn", "/etc/systemd/resolved.conf"
        )
        self.network_conf_dir = config.get(
            "network_conf_dir", "/etc/systemd/network/"
        )

    def generate_match_section(self, iface, cfg: CfgParser):
        sec = "Match"
        match_dict = {
            "name": "Name",
            "driver": "Driver",
        }

        if iface["type"] == "physical":
            match_dict["mac_address"] = "MACAddress"

        if not iface:
            return

        for k, v in match_dict.items():
            if k in iface and iface[k]:
                cfg.update_section(sec, v, iface[k])

        return iface["name"]

    def generate_link_section(self, iface, cfg: CfgParser):
        sec = "Link"

        if not iface:
            return

        if iface.get("mtu"):
            cfg.update_section(sec, "MTUBytes", iface["mtu"])

        if iface["type"] != "physical" and iface.get("mac_address"):
            cfg.update_section(sec, "MACAddress", iface["mac_address"])

        if "optional" in iface and iface["optional"]:
            cfg.update_section(sec, "RequiredForOnline", "no")

    def parse_routes(self, rid, conf, cfg: CfgParser):
        """
        Parse a route and use rid as a key in order to isolate the route from
        others in the route dict.
        """
        sec = "Route"
        route_cfg_map = {
            "gateway": "Gateway",
            "network": "Destination",
            "metric": "Metric",
            "table": "Table",
        }

        # prefix is derived using netmask by network_state
        prefix = ""
        if "prefix" in conf:
            prefix = f"/{conf['prefix']}"

        for k, v in conf.items():
            if k not in route_cfg_map:
                continue
            if k == "network":
                v += prefix
            cfg.update_route_section(sec, rid, route_cfg_map[k], v)

    def parse_subnets(self, iface, cfg: CfgParser):
        dhcp = "no"
        sec = "Network"
        rid = 0
        for e in iface.get("subnets", []):
            t = e["type"]
            if t in {"dhcp4", "dhcp"}:
                if dhcp == "no":
                    dhcp = "ipv4"
                elif dhcp == "ipv6":
                    dhcp = "yes"
            elif t == "dhcp6":
                if dhcp == "no":
                    dhcp = "ipv6"
                elif dhcp == "ipv4":
                    dhcp = "yes"
            if "routes" in e and e["routes"]:
                for i in e["routes"]:
                    # Use "r" as a dict key prefix for this route to isolate
                    # it from other sources of routes
                    self.parse_routes(f"r{rid}", i, cfg)
                    rid = rid + 1
            if "address" in e:
                addr = e["address"]
                if "prefix" in e:
                    addr += f"/{e['prefix']}"
                subnet_cfg_map = {
                    "address": "Address",
                    "gateway": "Gateway",
                    "dns_nameservers": "DNS",
                    "dns_search": "Domains",
                }
                for k, v in e.items():
                    if k == "address":
                        cfg.update_section("Address", subnet_cfg_map[k], addr)
                    elif k == "gateway":
                        # Use "a" as a dict key prefix for this route to
                        # isolate it from other sources of routes
                        cfg.update_route_section(
                            "Route", f"a{rid}", subnet_cfg_map[k], v
                        )
                        if should_add_gateway_onlink_flag(v, addr):
                            LOG.debug(
                                "Gateway %s is not contained within subnet %s,"
                                " adding GatewayOnLink flag",
                                v,
                                addr,
                            )
                            cfg.update_route_section(
                                "Route", f"a{rid}", "GatewayOnLink", "yes"
                            )
                        rid = rid + 1
                    elif k in {"dns_nameservers", "dns_search"}:
                        cfg.update_section(sec, subnet_cfg_map[k], " ".join(v))

        cfg.update_section(sec, "DHCP", dhcp)

        if isinstance(iface.get("accept-ra"), bool):
            val = "no"
            if iface["accept-ra"]:
                val = "yes"
            cfg.update_section(sec, "IPv6AcceptRA", val)

        return dhcp

    # This is to accommodate extra keys present in VMware config
    def dhcp_domain(self, d, cfg: CfgParser):
        for item in ["dhcp4domain", "dhcp6domain"]:
            if item not in d:
                continue
            ret = str(d[item]).casefold()
            try:
                ret = util.translate_bool(ret)
                ret = "yes" if ret else "no"
            except ValueError:
                if ret != "route":
                    LOG.warning("Invalid dhcp4domain value - %s", ret)
                    ret = "no"
            if item == "dhcp4domain":
                section = "DHCPv4"
            else:
                section = "DHCPv6"
            cfg.update_section(section, "UseDomains", ret)

    def parse_dns(self, iface, cfg: CfgParser, ns: NetworkState):
        sec = "Network"

        dns = iface.get("dns")
        if not dns and ns.version == 1:
            dns = {
                "search": ns.dns_searchdomains,
                "nameservers": ns.dns_nameservers,
            }
        elif not dns and ns.version == 2:
            return

        if dns.get("search"):
            cfg.update_section(sec, "Domains", " ".join(dns["search"]))
        if dns.get("nameservers"):
            cfg.update_section(sec, "DNS", " ".join(dns["nameservers"]))

    def parse_dhcp_overrides(self, cfg: CfgParser, device, dhcp, version):
        dhcp_config_maps = {
            "UseDNS": "use-dns",
            "UseDomains": "use-domains",
            "UseHostname": "use-hostname",
            "UseNTP": "use-ntp",
        }

        if version == "4":
            dhcp_config_maps.update(
                {
                    "SendHostname": "send-hostname",
                    "Hostname": "hostname",
                    "RouteMetric": "route-metric",
                    "UseMTU": "use-mtu",
                    "UseRoutes": "use-routes",
                }
            )

        if f"dhcp{version}-overrides" in device and dhcp in [
            "yes",
            f"ipv{version}",
        ]:
            dhcp_overrides = device[f"dhcp{version}-overrides"]
            for k, v in dhcp_config_maps.items():
                if v in dhcp_overrides:
                    cfg.update_section(f"DHCPv{version}", k, dhcp_overrides[v])

    def create_network_file(self, link, conf, nwk_dir, ext=".network"):
        net_fn_owner = "systemd-network"

        LOG.debug("Setting Networking Config for %s", link)
        net_fn = f"{nwk_dir}10-cloud-init-{link}{ext}"

        if conf.endswith("\n\n"):
            conf = conf[:-1]
        util.write_file(net_fn, conf)
        util.chownbyname(net_fn, net_fn_owner, net_fn_owner)

    def render_network_state(
        self,
        network_state: NetworkState,
        templates: Optional[dict] = None,
        target=None,
    ) -> None:
        network_dir = self.network_conf_dir
        if target:
            network_dir = subp.target_path(target) + network_dir

        util.ensure_dir(network_dir)

        network = self._render_content(network_state)
        vlan_netdev = network.pop("vlan_netdev", {})
        bond_netdev = network.pop("bond_netdev", {})
        bridge_netdev = network.pop("bridge_netdev", {})

        for k, v in network.items():
            self.create_network_file(k, v, network_dir)

        for k, v in vlan_netdev.items():
            self.create_network_file(k, v, network_dir, ext=".netdev")

        for k, v in bond_netdev.items():
            self.create_network_file(k, v, network_dir, ext=".netdev")

        for k, v in bridge_netdev.items():
            self.create_network_file(k, v, network_dir, ext=".netdev")

    def parseRoutingPolicy(self, policy, cfg: CfgParser):
        rid = 0
        key = "RoutingPolicyRule"
        for val in policy:
            for k, v in val.items():
                cfg.update_route_section(key, rid, k.capitalize(), v)
            rid += 1

    def extractRoutingPolicies(self, nsCfg: Dict):
        routingPolicies = {}
        key = "routing-policy"

        for section in ("ethernets", "bonds", "vlans", "bridges"):
            if section not in nsCfg:
                continue
            for iface, settings in nsCfg[section].items():
                if key in settings:
                    routingPolicies[iface] = settings[key]
        return routingPolicies

    def _render_content(self, ns: NetworkState):
        ret_dict = {}

        vlan_dict = self.render_vlans(ns)
        vlan_netdev = vlan_dict["vlan_netdev"]
        vlan_link = vlan_dict["vlan_link"]
        ret_dict["vlan_netdev"] = vlan_netdev

        bond_dict = self.render_bonds(ns)
        bond_netdev = bond_dict["bond_netdev"]
        bond_link = bond_dict["bond_link"]
        ret_dict["bond_netdev"] = bond_netdev

        bridge_dict = self.render_bridges(ns)
        bridge_netdev = bridge_dict["bridge_netdev"]
        bridge_link = bridge_dict["bridge_link"]
        ret_dict["bridge_netdev"] = bridge_netdev

        routingPolicies = self.extractRoutingPolicies(ns.config)

        for iface in ns.iter_interfaces():
            cfg = CfgParser()

            iface_name = iface["name"]

            vlan_link_name = vlan_link.get(iface_name, [])
            for i in vlan_link_name:
                cfg.update_section("Network", "VLAN", i)

            rPolicy = routingPolicies.get(iface_name)
            if rPolicy:
                self.parseRoutingPolicy(rPolicy, cfg)

            bond_link_name = bond_link.get(iface_name)
            if bond_link_name:
                cfg.update_section("Network", "Bond", bond_link_name)

            bridge_link_name = bridge_link.get(iface_name)
            if bridge_link_name:
                val = (
                    bridge_link["path-cost"]
                    .get(bridge_link_name, {})
                    .get(iface_name)
                )
                if val:
                    cfg.update_section("Bridge", "Cost", val)

                val = (
                    bridge_link["port-priority"]
                    .get(bridge_link_name, {})
                    .get(iface_name)
                )
                if val:
                    cfg.update_section("Bridge", "Priority", val)

                cfg.update_section("Network", "Bridge", bridge_link_name)

            link = self.generate_match_section(iface, cfg)

            self.generate_link_section(iface, cfg)
            dhcp = self.parse_subnets(iface, cfg)
            self.parse_dns(iface, cfg, ns)

            rid = 0
            for route in ns.iter_routes():
                # Use "c" as a dict key prefix for this route to isolate it
                # from other sources of routes
                self.parse_routes(f"c{rid}", route, cfg)
                rid = rid + 1

            if ns.version == 2:
                name: Optional[str] = iface["name"]
                # network state doesn't give dhcp domain info
                # using ns.config as a workaround here

                # Check to see if this interface matches against an interface
                # from the network state that specified a set-name directive.
                # If there is a device with a set-name directive and it has
                # set-name value that matches the current name, then update the
                # current name to the device's name. That will be the value in
                # the ns.config['ethernets'] dict below.
                for dev_name, dev_cfg in ns.config["ethernets"].items():
                    if "set-name" in dev_cfg:
                        if dev_cfg.get("set-name") == name:
                            name = dev_name
                            break
                if name in ns.config["ethernets"]:
                    device = ns.config["ethernets"][name]

                    # dhcp{version}domain are extra keys only present in
                    # VMware config
                    self.dhcp_domain(device, cfg)
                    for version in ["4", "6"]:
                        if (
                            f"dhcp{version}domain" in device
                            and "use-domains"
                            in device.get(f"dhcp{version}-overrides", {})
                        ):
                            exception = (
                                f"{name} has both dhcp{version}domain"
                                f" and dhcp{version}-overrides.use-domains"
                                f" configured. Use one"
                            )
                            raise RuntimeError(exception)

                        self.parse_dhcp_overrides(cfg, device, dhcp, version)

            ret_dict.update({link: cfg.get_final_conf()})

        return ret_dict

    def render_vlans(self, ns: NetworkState) -> dict:
        vlan_link_info = defaultdict(list)
        vlan_ndev_configs = {}

        vlan_filter = renderer.filter_by_type("vlan")
        for iface in ns.iter_interfaces(vlan_filter):
            iface_name = iface["name"]
            vlan_id = iface.get("vlan_id")
            parent = iface.get("vlan-raw-device")

            if vlan_id is None or parent is None:
                LOG.warning(
                    "Skipping VLAN %s - missing 'id' or 'link'", iface_name
                )
                continue

            vlan_link_info[parent].append(iface_name)

            # -------- .netdev for VLAN --------
            cfg = CfgParser()
            cfg.update_section("NetDev", "Name", iface_name)
            cfg.update_section("NetDev", "Kind", "vlan")

            val = iface.get("mtu")
            if val:
                cfg.update_section("NetDev", "MTUBytes", val)

            val = iface.get("mac_address")
            if val:
                cfg.update_section("NetDev", "MACAddress", val.lower())

            cfg.update_section("VLAN", "Id", vlan_id)
            vlan_ndev_configs[iface_name] = cfg.get_final_conf()

        return {
            "vlan_netdev": vlan_ndev_configs,
            "vlan_link": vlan_link_info,
        }

    def render_bonds(self, ns: NetworkState) -> dict:
        bond_link_info: Dict[str, Any] = {}
        bond_ndev_configs = {}
        section = "Bond"

        bond_filter = renderer.filter_by_type("bond")
        slave_filter = renderer.filter_by_attr("bond-master")
        for iface in ns.iter_interfaces(bond_filter):
            iface_name = iface["name"]
            bond_slaves = [
                slave_iface["name"]
                for slave_iface in ns.iter_interfaces(slave_filter)
                if slave_iface["bond-master"] == iface_name
            ]
            if not bond_slaves:
                LOG.warning(
                    "Skipping bond %s - missing 'interfaces'", iface_name
                )
                continue

            bond_link_info.update({iface: iface_name for iface in bond_slaves})

            # -------- .netdev for Bond --------
            cfg = CfgParser()
            cfg.update_section("NetDev", "Name", iface_name)
            cfg.update_section("NetDev", "Kind", "bond")

            val = iface.get("mtu")
            if val:
                cfg.update_section("NetDev", "MTUBytes", val)

            val = iface.get("mac_address")
            if val:
                cfg.update_section("NetDev", "MACAddress", val.lower())

            # Optional bond parameters
            ParamMapType = Dict[str, Tuple[str, Callable[[object], str]]]

            param_map: ParamMapType = {
                "bond-ad-select": ("AdSelect", str),
                "bond-arp-interval": ("ARPIntervalSec", lambda v: f"{v}ms"),
                "bond-arp-validate": ("ARPValidate", str),
                "bond-downdelay": ("DownDelaySec", lambda v: f"{v}ms"),
                "bond-fail-over-mac": ("FailOverMACPolicy", str),
                "bond-lacp-rate": ("LACPTransmitRate", str),
                "bond-miimon": ("MIIMonitorSec", lambda v: f"{v}ms"),
                "bond-min-links": ("MinLinks", str),
                "bond-mode": ("Mode", str),
                "bond-num-grat-arp": ("GratuitousARP", str),
                "bond-primary-reselect": ("PrimaryReselectPolicy", str),
                "bond-updelay": ("UpDelaySec", lambda v: f"{v}ms"),
                "bond-xmit-hash-policy": ("TransmitHashPolicy", str),
                "bond-all-slaves-active": (
                    "AllSlavesActive",
                    lambda v: str(v).lower(),
                ),
                "bond-arp-all-targets": ("ARPAllTargets", str),
                "bond-learn-packet-interval": ("LearnPacketIntervalSec", str),
                "bond-packets-per-slave": ("PacketsPerSlave", str),
            }

            for key, (option, formatter) in param_map.items():
                if key in iface:
                    cfg.update_section(section, option, formatter(iface[key]))

            if "arp-ip-targets" in iface:
                targets = iface["arp-ip-targets"]
                if isinstance(targets, str):
                    targets = [targets]
                cfg.update_section(section, "ARPIPTargets", " ".join(targets))

            bond_ndev_configs[iface_name] = cfg.get_final_conf()

        return {
            "bond_netdev": bond_ndev_configs,
            "bond_link": bond_link_info,
        }

    def render_bridges(self, ns: NetworkState) -> dict:
        bridge_link_info: Dict[str, Any] = {
            "macaddress": {},
            "path-cost": {},
            "port-priority": {},
        }

        bridge_ndev_configs = {}

        bridge_filter = renderer.filter_by_type("bridge")
        for iface in ns.iter_interfaces(bridge_filter):
            iface_name = iface["name"]
            bridge_ports = iface.get("bridge_ports", [])
            if not bridge_ports:
                LOG.warning(
                    "Skipping bridge %s - missing 'interfaces'", iface_name
                )
                continue

            # Map each interface to its bridge
            for bridge_port in bridge_ports:
                bridge_link_info[bridge_port] = iface_name

            # ---- .netdev config for the bridge ----
            cfg = CfgParser()
            cfg.update_section("NetDev", "Name", iface_name)
            cfg.update_section("NetDev", "Kind", "bridge")

            val = iface.get("mtu")
            if val:
                cfg.update_section("NetDev", "MTUBytes", val)

            val = iface.get("mac_address")
            if val:
                cfg.update_section("NetDev", "MACAddress", val.lower())

            # Bridge parameters
            param_map = {
                "bridge_ageing": "AgeingTimeSec",
                "bridge_bridgeprio": "Priority",
                "bridge_fd": "ForwardDelaySec",
                "bridge_hello": "HelloTimeSec",
                "bridge_maxage": "MaxAgeSec",
                "bridge_stp": "STP",
            }
            for key, sysd_key in param_map.items():
                val = iface.get(key)
                if val:
                    if isinstance(val, bool):
                        val = "yes" if val else "no"
                    cfg.update_section("Bridge", sysd_key, val)

            val = iface.get("bridge_pathcost")
            if val:
                bridge_link_info["path-cost"][iface_name] = val

            val = iface.get("bridge_portprio")
            if val:
                bridge_link_info["port-priority"][iface_name] = val

            bridge_ndev_configs[iface_name] = cfg.get_final_conf()

        return {
            "bridge_netdev": bridge_ndev_configs,
            "bridge_link": bridge_link_info,
        }


def available(target=None):
    expected = ["ip", "systemctl"]
    search = ["/usr/sbin", "/bin"]
    for p in expected:
        if not subp.which(p, search=search):
            return False
    return True



# This file is part of cloud-init. See LICENSE file for license information.

import re

import pytest

from cloudinit.distros.parsers import resolv_conf

BASE_RESOLVE = """
; generated by /sbin/dhclient-script
search blah.yahoo.com yahoo.com
nameserver 10.15.44.14
nameserver 10.15.30.92
"""
BASE_RESOLVE = BASE_RESOLVE.strip()


class TestResolvHelper:
    def test_parse_same(self):
        rp = resolv_conf.ResolvConf(BASE_RESOLVE)
        rp_r = str(rp).strip()
        assert BASE_RESOLVE == rp_r

    def test_local_domain(self):
        rp = resolv_conf.ResolvConf(BASE_RESOLVE)
        assert rp.local_domain is None

        rp.local_domain = "bob"
        assert "bob" == rp.local_domain
        assert "domain bob" in str(rp)

    def test_nameservers(self):
        rp = resolv_conf.ResolvConf(BASE_RESOLVE)

        # Start with two nameservers that already appear in the configuration.
        assert "10.15.44.14" in rp.nameservers
        assert "10.15.30.92" in rp.nameservers

        # Add a third nameserver and verify it appears in the resolv.conf.
        rp.add_nameserver("10.2")
        assert "10.2" in rp.nameservers
        assert "nameserver 10.2" in str(rp)
        assert len(rp.nameservers) == 3

        # Add a fourth nameserver and verify it appears in the resolv.conf.
        rp.add_nameserver("10.3")
        assert "10.3" in rp.nameservers
        assert "nameserver 10.3" in str(rp)
        assert len(rp.nameservers) == 4

    def test_search_domains(self):
        rp = resolv_conf.ResolvConf(BASE_RESOLVE)
        assert "yahoo.com" in rp.search_domains
        assert "blah.yahoo.com" in rp.search_domains
        rp.add_search_domain("bbb.y.com")
        assert "bbb.y.com" in rp.search_domains
        assert re.search(r"search(.*)bbb.y.com(.*)", str(rp))
        assert "bbb.y.com" in rp.search_domains
        rp.add_search_domain("bbb.y.com")
        assert len(rp.search_domains) == 3
        rp.add_search_domain("bbb2.y.com")
        assert len(rp.search_domains) == 4
        rp.add_search_domain("bbb3.y.com")
        assert len(rp.search_domains) == 5
        rp.add_search_domain("bbb4.y.com")
        assert len(rp.search_domains) == 6
        with pytest.raises(ValueError):
            rp.add_search_domain("bbb5.y.com")
        assert len(rp.search_domains) == 6



# Author: Jonas Keidel <jonas.keidel@hetzner.com>
# Author: Markus Schade <markus.schade@hetzner.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

from typing import Optional, Tuple

from cloudinit import net, url_helper


def _skip_retry_on_empty_response(cause: url_helper.UrlError) -> bool:
    return cause.code != 204


def get_metadata(
    urls,
    max_wait=120,
    timeout=2,
    sleep_time=2,
) -> Tuple[Optional[str], bytes]:
    try:
        url, contents = url_helper.wait_for_url(
            urls=urls,
            max_wait=max_wait,
            timeout=timeout,
            sleep_time=sleep_time,
            # It is ok for userdata to not exist (that's why we are stopping if
            # HTTP code is 204) and just in that case returning an empty
            # string.
            exception_cb=_skip_retry_on_empty_response,
        )
        if not url:
            raise RuntimeError("No data received from urls: '%s':" % urls)
        return url, contents
    except url_helper.UrlError as e:
        if e.code == 204:
            return e.url, b""
        raise


def get_interface_name_from_mac(mac: str) -> Optional[str]:
    mac_to_iface = net.get_interfaces_by_mac()
    return mac_to_iface.get(mac.lower())
